import pkg_resources
import unittest


class CodeStyleTestCase(unittest.TestCase):
    def test_code_style(self):
        flake8 = pkg_resources.load_entry_point('flake8', 'console_scripts', 'flake8')
        try:
            flake8([])
        except SystemExit as e:
            if e.code != 0:
                self.fail('Code style checks failed')



import datetime
from unittest import TestCase
from collections import OrderedDict

from bai2.helpers import IteratorHelper
from bai2.parsers import TransactionDetailParser, AccountParser, \
    GroupParser, Bai2FileParser
from bai2.constants import TypeCodes, FundsType, GroupStatus, \
    AsOfDateModifier
from bai2.models import \
    Bai2File, Bai2FileHeader, Bai2FileTrailer, \
    Group, GroupHeader, GroupTrailer, \
    Account, AccountIdentifier, AccountTrailer, \
    TransactionDetail
from bai2.exceptions import ParsingException, \
    NotSupportedYetException, IntegrityException


class TransactionDetailParserTestCase(TestCase):
    def test_parse(self):
        lines = [
            '16,165,1500000,0,DD1620,,DEALER PAYMENTS',
        ]

        ii = IteratorHelper(lines)
        parser = TransactionDetailParser(ii)

        transaction = parser.parse()

        self.assertEqual(transaction.type_code, TypeCodes['165'])
        self.assertEqual(transaction.amount, 1500000)
        self.assertEqual(transaction.funds_type, FundsType.immediate_availability)
        self.assertEqual(transaction.bank_reference, 'DD1620')
        self.assertEqual(transaction.customer_reference, None)
        self.assertEqual(
            transaction.text,
            'DEALER PAYMENTS'
        )

    def test_continuation_record(self):
        lines = [
            '16,115,10000000,S,5000000,4000000,1000000/',
            '88,AX13612,B096132,AMALGAMATED CORP. LOCKBOX',
            '88,DEPOSIT-MISC. RECEIVABLES'
        ]

        parser = TransactionDetailParser(IteratorHelper(lines))

        transaction = parser.parse()

        self.assertEqual(transaction.type_code, TypeCodes['115'])
        self.assertEqual(transaction.amount, 10000000)
        self.assertEqual(transaction.funds_type, FundsType.distributed_availability_simple)
        self.assertEqual(
            transaction.availability,
            OrderedDict([('0', 5000000), ('1', 4000000), ('>1', 1000000)])
        )
        self.assertEqual(transaction.bank_reference, 'AX13612')
        self.assertEqual(transaction.customer_reference, 'B096132')
        self.assertEqual(
            transaction.text,
            'AMALGAMATED CORP. LOCKBOX DEPOSIT-MISC. RECEIVABLES'
        )

    def test_unknown_availability(self):
        lines = [
            '16,165,1500000,Z,DD1620,,DEALER PAYMENTS',
        ]

        ii = IteratorHelper(lines)
        parser = TransactionDetailParser(ii)

        transaction = parser.parse()

        self.assertEqual(transaction.type_code, TypeCodes['165'])
        self.assertEqual(transaction.amount, 1500000)
        self.assertEqual(transaction.funds_type, FundsType.unknown_availability)
        self.assertEqual(transaction.availability, None)
        self.assertEqual(transaction.bank_reference, 'DD1620')
        self.assertEqual(transaction.customer_reference, None)
        self.assertEqual(
            transaction.text,
            'DEALER PAYMENTS'
        )

    def test_value_dated_availability(self):
        lines = [
            '16,191,001,V,150715,2340,1234567890,RP12312312312312/',
            '88,FR:FP SIP INCOMING',
            '88,ENDT:20150715',
            '88,TRID:RP12312312312312',
            '88,PY:RP1231231231231200                 A1234BC 22/03/66',
            '88,BI:22222222',
            '88,OB:111111 BUCKINGHAM PALACE OB3:BARCLAYS BANK PLC',
            '88,BO:11111111 BO1:DOE JO',
        ]

        parser = TransactionDetailParser(IteratorHelper(lines))

        transaction = parser.parse()

        self.assertEqual(transaction.funds_type, FundsType.value_dated)
        self.assertEqual(
            transaction.availability['date'],
            datetime.date(day=15, month=7, year=2015)
        )
        self.assertEqual(
            transaction.availability['time'],
            datetime.time(hour=23, minute=40)
        )

    def test_distributed_availability_simple(self):
        lines = [
            '16,191,005,S,001,003,001,1234567890,RP12312312312312/',
            '88,FR:FP SIP INCOMING',
            '88,ENDT:20150715',
            '88,TRID:RP12312312312312',
            '88,PY:RP1231231231231200                 A1234BC 22/03/66',
            '88,BI:22222222',
            '88,OB:111111 BUCKINGHAM PALACE OB3:BARCLAYS BANK PLC',
            '88,BO:11111111 BO1:DOE JO',
        ]

        parser = TransactionDetailParser(IteratorHelper(lines))

        transaction = parser.parse()

        self.assertEqual(transaction.funds_type, FundsType.distributed_availability_simple)
        self.assertEqual(transaction.amount, 5)
        self.assertEqual(
            transaction.availability,
            OrderedDict([('0', 1), ('1', 3), ('>1', 1)])
        )

    def test_distributed_availability(self):
        lines = [
            '16,191,005,D,2,1,001,2,004,1234567890,RP12312312312312/',
            '88,FR:FP SIP INCOMING',
            '88,ENDT:20150715',
            '88,TRID:RP12312312312312',
            '88,PY:RP1231231231231200                 A1234BC 22/03/66',
            '88,BI:22222222',
            '88,OB:111111 BUCKINGHAM PALACE OB3:BARCLAYS BANK PLC',
            '88,BO:11111111 BO1:DOE JO',
        ]

        parser = TransactionDetailParser(IteratorHelper(lines))

        transaction = parser.parse()

        self.assertEqual(transaction.funds_type, FundsType.distributed_availability)
        self.assertEqual(transaction.amount, 5)

        self.assertEqual(
            transaction.availability,
            OrderedDict([('1', 1), ('2', 4)])
        )


class AccountParserTestCase(TestCase):

    def test_parse(self):
        lines = [
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '49,72000000,3/'
        ]

        parser = AccountParser(IteratorHelper(lines))

        account = parser.parse()

        header = account.header
        self.assertEqual(header.customer_account_number, '0975312468')
        self.assertEqual(header.currency, 'GBP')

        trailer = account.trailer
        self.assertEqual(trailer.account_control_total, 72000000)
        self.assertEqual(trailer.number_of_records, 3)

        self.assertEqual(len(account.children), 1)

    def test_parse_without_transactions(self):
        lines = [
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '49,70500000,2/'
        ]

        parser = AccountParser(IteratorHelper(lines))

        account = parser.parse()

        self.assertEqual(len(account.children), 0)

    def test_parse_with_multiple_transactions(self):
        lines = [
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '16,115,10000000,S,5000000,4000000,1000000/',
            '88,AX13612,B096132,AMALGAMATED CORP. LOCKBOX',
            '88,DEPOSIT-MISC. RECEIVABLES',
            '49,82000000,6/'
        ]

        parser = AccountParser(IteratorHelper(lines))

        account = parser.parse()

        self.assertEqual(len(account.children), 2)

    def test_fails_without_trailer(self):
        lines = [
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
        ]

        parser = AccountParser(IteratorHelper(lines))
        self.assertRaises(ParsingException, parser.parse)

    def test_fails_integrity_on_numbers_of_records(self):
        """
        Number of records == 4 when it should be 3.
        """
        lines = [
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '49,72000000,4/'
        ]

        parser = AccountParser(IteratorHelper(lines))
        self.assertRaises(IntegrityException, parser.parse)

    def test_fails_integrity_on_account_control_total(self):
        """
        Account Control Total == 72000001 when it should be 72000000.
        """
        lines = [
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '49,72000001,3/'
        ]

        parser = AccountParser(IteratorHelper(lines))
        self.assertRaises(IntegrityException, parser.parse)

    def test_ignore_integrity_checks(self):
        """
        Checks that if IGNORE_INTEGRITY_CHECKS is set, integrity checks
        are not performed.
        """
        lines = [
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '49,72000001,3/'
        ]

        parser = AccountParser(IteratorHelper(lines), check_integrity=False)
        account = parser.parse()
        self.assertTrue(isinstance(account, Account))


class GroupParserTestCase(TestCase):

    def test_parse(self):
        lines = [
            '02,031001234,122099999,1,040620,2359,GBP,2/',
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '49,72000000,3/',
            '98,72000000,1,5/'
        ]

        parser = GroupParser(IteratorHelper(lines))

        group = parser.parse()

        header = group.header
        self.assertEqual(header.ultimate_receiver_id, '031001234')
        self.assertEqual(header.originator_id, '122099999')
        self.assertEqual(header.group_status, GroupStatus.update)
        self.assertEqual(header.as_of_date, datetime.date(year=2004, month=6, day=20))
        self.assertEqual(header.as_of_time, datetime.time(hour=23, minute=59))
        self.assertEqual(header.currency, 'GBP')
        self.assertEqual(header.as_of_date_modifier, AsOfDateModifier.final_previous_day)

        trailer = group.trailer
        self.assertEqual(trailer.group_control_total, 72000000)
        self.assertEqual(trailer.number_of_accounts, 1)
        self.assertEqual(trailer.number_of_records, 5)

        self.assertEqual(len(group.children), 1)

    def test_default_currency(self):
        """
        If currency field is blank in group header => default to USD
        """

        lines = [
            '02,031001234,122099999,1,040620,2359,,2/',
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '49,72000000,3/',
            '98,72000000,1,5/'
        ]

        parser = GroupParser(IteratorHelper(lines))

        group = parser.parse()

        header = group.header
        self.assertEqual(header.currency, 'USD')

    def test_multiple_accounts(self):
        lines = [
            '02,031001234,122099999,1,040620,2359,,2/',
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '49,72000000,3/',
            '03,0975312469,GBP,010,100,,/',
            '49,100,2/',
            '98,72000100,2,7/'
        ]

        parser = GroupParser(IteratorHelper(lines))

        group = parser.parse()

        trailer = group.trailer
        self.assertEqual(trailer.group_control_total, 72000100)
        self.assertEqual(trailer.number_of_accounts, 2)
        self.assertEqual(trailer.number_of_records, 7)

        self.assertEqual(len(group.children), 2)

    def test_fails_if_no_accounts_found(self):
        lines = [
            '02,031001234,122099999,1,040620,2359,,2/',
            '98,11800000,0,2/'
        ]

        parser = GroupParser(IteratorHelper(lines))
        self.assertRaises(ParsingException, parser.parse)

    def test_fails_without_trailer(self):
        lines = [
            '02,031001234,122099999,1,040620,2359,,2/',
        ]

        parser = GroupParser(IteratorHelper(lines))
        self.assertRaises(ParsingException, parser.parse)

    def test_fails_integrity_on_numbers_of_records(self):
        """
        Number of records == 6 when it should be 5.
        """
        lines = [
            '02,031001234,122099999,1,040620,2359,GBP,2/',
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '49,72000000,3/',
            '98,72000000,1,6/'
        ]

        parser = GroupParser(IteratorHelper(lines))
        self.assertRaises(IntegrityException, parser.parse)

    def test_fails_integrity_on_numbers_of_accounts(self):
        """
        Number of accounts == 4 when it should be 3.
        """
        lines = [
            '02,031001234,122099999,1,040620,2359,GBP,2/',
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '49,72000000,4/',
            '98,72000000,1,5/'
        ]

        parser = GroupParser(IteratorHelper(lines))
        self.assertRaises(IntegrityException, parser.parse)

    def test_fails_integrity_on_group_control_total(self):
        """
        Group Control Total == 72000001 when it should be 72000000.
        """
        lines = [
            '02,031001234,122099999,1,040620,2359,GBP,2/',
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '49,72000000,3/',
            '98,72000001,1,5/'
        ]

        parser = GroupParser(IteratorHelper(lines))
        self.assertRaises(IntegrityException, parser.parse)

    def test_ignore_integrity_checks(self):
        """
        Checks that if IGNORE_INTEGRITY_CHECKS is set, integrity checks
        are not performed.
        """
        lines = [
            '02,031001234,122099999,1,040620,2359,GBP,2/',
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '49,72000000,3/',
            '98,72000001,2,6/'
        ]

        parser = GroupParser(IteratorHelper(lines), check_integrity=False)
        group = parser.parse()
        self.assertTrue(isinstance(group, Group))


class Bai2FileParserTestCase(TestCase):

    def test_parse(self):
        lines = [
            '01,CITIDIRECT,8888888,150716,0713,00131100,,,2/',
            '02,8888888,CITIGB00,1,150715,2340,GBP,2/',
            '03,77777777,GBP,010,10000,,,015,10000,,,/',
            '16,191,001,V,150715,,1234567890,RP12312312312312/',
            '88,FR:FP SIP INCOMING',
            '88,ENDT:20150715',
            '88,TRID:RP12312312312312',
            '88,PY:RP1231231231231200                 A1234BC 22/03/66',
            '88,BI:22222222',
            '88,OB:111111 BUCKINGHAM PALACE OB3:BARCLAYS BANK PLC',
            '88,BO:11111111 BO1:DOE JO',
            '49,20001,10/',
            '98,20001,1,12/',
            '99,20001,1,14/'
        ]

        parser = Bai2FileParser(IteratorHelper(lines))
        bai2_file = parser.parse()

        july_15_2015 = datetime.date(day=15, month=7, year=2015)
        july_16_2015 = datetime.date(day=16, month=7, year=2015)

        # BAI2 file
        self.assertTrue(isinstance(bai2_file, Bai2File))

        # BAI2 file header
        bai2_header = bai2_file.header
        self.assertTrue(isinstance(bai2_header, Bai2FileHeader))
        self.assertEqual(bai2_header.sender_id, 'CITIDIRECT')
        self.assertEqual(bai2_header.receiver_id, '8888888')
        self.assertEqual(bai2_header.creation_date, july_16_2015)
        self.assertEqual(bai2_header.creation_time, datetime.time(hour=7, minute=13))
        self.assertEqual(bai2_header.file_id, '00131100')
        self.assertEqual(bai2_header.physical_record_length, None)
        self.assertEqual(bai2_header.block_size, None)
        self.assertEqual(bai2_header.version_number, 2)

        # BAI2 file trailer
        bai2_trailer = bai2_file.trailer
        self.assertTrue(isinstance(bai2_trailer, Bai2FileTrailer))
        self.assertEqual(bai2_trailer.file_control_total, 20001)
        self.assertEqual(bai2_trailer.number_of_groups, 1)
        self.assertEqual(bai2_trailer.number_of_records, 14)

        # GROUP

        self.assertEqual(len(bai2_file.children), 1)
        group = bai2_file.children[0]
        self.assertTrue(isinstance(group, Group))

        # Group Header

        group_header = group.header
        self.assertTrue(isinstance(group_header, GroupHeader))
        self.assertEqual(group_header.ultimate_receiver_id, '8888888')
        self.assertEqual(group_header.originator_id, 'CITIGB00')
        self.assertEqual(group_header.group_status, GroupStatus.update)
        self.assertEqual(group_header.as_of_date, july_15_2015)
        self.assertEqual(group_header.as_of_time, datetime.time(hour=23, minute=40))
        self.assertEqual(group_header.currency, 'GBP')
        self.assertEqual(
            group_header.as_of_date_modifier, AsOfDateModifier.final_previous_day
        )

        # Group Trailer

        group_trailer = group.trailer
        self.assertTrue(isinstance(group_trailer, GroupTrailer))
        self.assertEqual(group_trailer.group_control_total, 20001)
        self.assertEqual(group_trailer.number_of_accounts, 1)
        self.assertEqual(group_trailer.number_of_records, 12)

        # ACCOUNT

        self.assertEqual(len(group.children), 1)
        account = group.children[0]
        self.assertTrue(isinstance(account, Account))

        # Account Identifier

        account_identifier = account.header
        self.assertTrue(isinstance(account_identifier, AccountIdentifier))
        self.assertEqual(account_identifier.customer_account_number, '77777777')
        self.assertEqual(account_identifier.currency, 'GBP')

        # Account Trailer

        account_trailer = account.trailer
        self.assertTrue(isinstance(account_trailer, AccountTrailer))
        self.assertEqual(account_trailer.account_control_total, 20001)
        self.assertEqual(account_trailer.number_of_records, 10)

        # Transaction Detail

        self.assertEqual(len(account.children), 1)
        transaction = account.children[0]
        self.assertTrue(isinstance(transaction, TransactionDetail))
        self.assertEqual(transaction.type_code, TypeCodes['191'])
        self.assertEqual(transaction.amount, 1)
        self.assertEqual(transaction.funds_type, FundsType.value_dated)
        self.assertEqual(transaction.availability['date'], july_15_2015)
        self.assertEqual(transaction.availability['time'], None)
        self.assertEqual(transaction.bank_reference, '1234567890')
        self.assertEqual(transaction.customer_reference, 'RP12312312312312')
        self.assertEqual(
            transaction.text,
            'FR:FP SIP INCOMING '
            'ENDT:20150715 '
            'TRID:RP12312312312312 '
            'PY:RP1231231231231200                 A1234BC 22/03/66 '
            'BI:22222222 '
            'OB:111111 BUCKINGHAM PALACE OB3:BARCLAYS BANK PLC '
            'BO:11111111 BO1:DOE JO'
        )

    def test_only_version_2_supported(self):
        """
        Checks that BAI version 2 is supported.
        """
        lines = [
            '01,122099999,123456789,040621,0200,1,,,1/',
            '02,031001234,122099999,1,040620,2359,GBP,2/',
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '49,18650000,3/',
            '98,18650000,1,5/',
            '99,18650000,1,7/'
        ]

        parser = Bai2FileParser(IteratorHelper(lines))

        self.assertRaises(NotSupportedYetException, parser.parse)

    def test_multiple_groups(self):
        lines = [
            '01,122099999,123456789,040621,0200,1,,,2/',
            '02,031001234,122099999,1,040620,2359,GBP,2/',
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '49,72000000,3/',
            '98,72000000,1,5/',
            '02,031001233,122099998,1,040620,2359,GBP,2/',
            '03,0975312469,GBP,010,100,,/',
            '49,100,2/',
            '98,100,1,4/',
            '99,72000100,2,11/'
        ]

        parser = Bai2FileParser(IteratorHelper(lines))
        bai2_file = parser.parse()

        trailer = bai2_file.trailer
        self.assertEqual(trailer.file_control_total, 72000100)
        self.assertEqual(trailer.number_of_groups, 2)
        self.assertEqual(trailer.number_of_records, 11)

        self.assertEqual(len(bai2_file.children), 2)

    def test_fails_if_no_groups_found(self):
        lines = [
            '01,122099999,123456789,040621,0200,1,,,2/',
            '99,1215450000,0,2/'
        ]

        parser = Bai2FileParser(IteratorHelper(lines))
        self.assertRaises(ParsingException, parser.parse)

    def test_fails_without_trailer(self):
        lines = [
            '01,122099999,123456789,040621,0200,1,,,2/',
        ]

        parser = Bai2FileParser(IteratorHelper(lines))
        self.assertRaises(ParsingException, parser.parse)

    def test_fails_integrity_on_numbers_of_records(self):
        """
        Number of records == 8 when it should be 7.
        """
        lines = [
            '01,122099999,123456789,040621,0200,1,,,2/',
            '02,031001234,122099999,1,040620,2359,GBP,2/',
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '49,72000000,3/',
            '98,72000000,1,5/',
            '99,72000000,1,8/'
        ]

        parser = Bai2FileParser(IteratorHelper(lines))
        self.assertRaises(IntegrityException, parser.parse)

    def test_fails_integrity_on_numbers_of_groups(self):
        """
        Number of groups == 2 when it should be 1.
        """
        lines = [
            '01,122099999,123456789,040621,0200,1,,,2/',
            '02,031001234,122099999,1,040620,2359,GBP,2/',
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '49,72000000,3/',
            '98,72000000,1,5/',
            '99,72000000,2,7/'
        ]

        parser = Bai2FileParser(IteratorHelper(lines))
        self.assertRaises(IntegrityException, parser.parse)

    def test_fails_integrity_on_file_control_total(self):
        """
        File Control Total == 72000001 when it should be 72000000.
        """
        lines = [
            '01,122099999,123456789,040621,0200,1,,,2/',
            '02,031001234,122099999,1,040620,2359,GBP,2/',
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '49,72000000,3/',
            '98,72000000,1,5/',
            '99,72000001,1,7/'
        ]

        parser = Bai2FileParser(IteratorHelper(lines))
        self.assertRaises(IntegrityException, parser.parse)

    def test_ignore_integrity_checks(self):
        """
        Checks that if IGNORE_INTEGRITY_CHECKS is set, integrity checks
        are not performed.
        """
        lines = [
            '01,122099999,123456789,040621,0200,1,,,2/',
            '02,031001234,122099999,1,040620,2359,GBP,2/',
            '03,0975312468,GBP,010,500000,,,190,70000000,4,0/',
            '16,165,1500000,1,DD1620,, DEALER PAYMENTS',
            '49,72000000,3/',
            '98,72000000,1,5/',
            '99,72000001,2,8/'
        ]

        parser = Bai2FileParser(IteratorHelper(lines), check_integrity=False)
        bai2_file = parser.parse()
        self.assertTrue(isinstance(bai2_file, Bai2File))



import datetime
import re

from .constants import TypeCodes


def parse_date(value):
    """
    YYMMDD Format.
    """
    return datetime.datetime.strptime(value, '%y%m%d').date()


def write_date(date):
    return date.strftime('%y%m%d')


def parse_time(value):
    clock_pattern = re.compile('\d\d:\d\d:\d\d')

    if clock_pattern.match(value):
        return parse_clock_time(value)
    else:
        return parse_military_time(value)


def parse_clock_time(value):
    return datetime.datetime.strptime(value, '%H:%M:%S').time()


def parse_military_time(value):
    """
    Military Format, 24 hours. 0001 through 2400.
    Times are stated in military format (0000 through 2400).
    0000 indicates the beginning of the day and 2400 indicates the end of the day
    for the date indicated.
    Some processors use 9999 to indicate the end of the day.
    Be prepared to recognize 9999 as end-of-day when receiving transmissions.
    """
    # 9999 indicates end of the day
    # 2400 indicates end of the day but 24:00 not allowed so
    # it's really 23:59
    if value == '9999' or value == '2400':
        return datetime.time.max
    return datetime.datetime.strptime(value, '%H%M').time()


def write_time(time, clock_format_for_intra_day=False):
    if clock_format_for_intra_day and time != datetime.time.max:
        return write_clock_time(time)
    else:
        return write_military_time(time)


def write_clock_time(time):
    date = datetime.datetime.now().replace(hour=time.hour,
                                           minute=time.minute,
                                           second=time.second)
    return datetime.datetime.strftime(date, '%H:%M:%S')


def write_military_time(time):
    if time == datetime.time.max:
        return '2400'
    else:
        date = datetime.datetime.now().replace(hour=time.hour, minute=time.minute)
        return datetime.datetime.strftime(date, '%H%M')


def parse_type_code(value):
    return TypeCodes[value]


def convert_to_string(value):
    if value is None:
        return ''
    else:
        return str(value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import importlib
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open('README.rst') as readme_file:
    readme = readme_file.read().strip()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '').strip()

package_info = importlib.import_module('bai2')

install_requires = []
tests_require = ['flake8']
if sys.version_info[0:2] < (3, 4):
    install_requires.append('enum34')

setup(
    name='bai2',
    version=package_info.__version__,
    author=package_info.__author__,
    author_email=package_info.__email__,
    url='https://github.com/ministryofjustice/bai2',
    license='MIT',
    description='BAI2 Parser',
    long_description=readme + '\n\n' + history,
    packages=['bai2'],
    package_dir={'bai2': 'bai2'},
    include_package_data=True,
    keywords='bai2 bookkeeping cash management balance reporting',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    install_requires=install_requires,
    tests_require=tests_require,
    test_suite='tests',
)



from collections import OrderedDict

from .models import \
    Bai2File, Bai2FileHeader, Bai2FileTrailer, \
    Group, GroupHeader, GroupTrailer, \
    AccountIdentifier, AccountTrailer, Account, \
    TransactionDetail
from .utils import write_date, write_time, convert_to_string
from .constants import CONTINUATION_CODE


class BaseWriter(object):

    def __init__(self, obj,
                 line_length=80,
                 text_on_new_line=False,
                 clock_format_for_intra_day=False):
        """
        Keyword arguments:
        line_length -- number of characters per record (default 80)
        text_on_new_line -- whether to begin a text field in a new record (default False)
        clock_format_for_intra_day -- use HH:MM:SS instead of HHMM for intra-day times (default False)
        """
        self.obj = obj
        self.line_length = line_length
        self.text_on_new_line = text_on_new_line
        self.clock_format_for_intra_day = clock_format_for_intra_day

    def write(self):
        raise NotImplementedError()


class BaseSectionWriter(BaseWriter):
    model = None
    header_writer_class = None
    child_writer_class = None
    trailer_writer_class = None

    def write(self):
        header = self.header_writer_class(self.obj.header).write()

        children = []
        for child in self.obj.children:
            children += self.child_writer_class(child).write()

        self.obj.update_totals()
        self.obj.trailer.number_of_records = len(header) + len(children) + 1
        trailer = self.trailer_writer_class(self.obj.trailer).write()

        return header + children + trailer


class BaseSingleWriter(BaseWriter):
    model = None

    def _write_field_from_config(self, field_config):
        if isinstance(field_config, str):
            field_config = (field_config, lambda w, x: x)

        field_name, write_func = field_config
        field_value = getattr(self.obj, field_name, None)
        output = write_func(self, field_value) if field_value is not None else None

        if isinstance(output, dict):
            return output
        else:
            return {field_name: convert_to_string(output)}

    def _write_fields_from_config(self, fields_config):
        fields = OrderedDict()
        for field_config in fields_config:
            fields.update(self._write_field_from_config(field_config))
        return fields

    def write(self):
        record = ''
        fields = self._write_fields_from_config(self.fields_config)

        record += self.model.code.value
        for field_name in fields:
            record += ',' + fields[field_name]
        record += '/'

        return [record]


def expand_availability(writer, availability):
    fields = OrderedDict()
    if len(availability) == 0:
        pass
    elif list(availability.keys()) in [['0', '1', '>1'], ['date', 'time']]:
        for field, value in availability.items():
            if field == 'date':
                value = write_date(value) if value else None
            elif field == 'time':
                value = (write_time(value, writer.clock_format_for_intra_day)
                         if value else None)
            fields[field] = convert_to_string(value)
    else:
        fields['distribution_length'] = str(len(availability))
        for field, value in availability.items():
            fields['day_%s' % str(field)] = convert_to_string(field)
            fields['amount_%s' % str(field)] = convert_to_string(value)
    return fields


class TransactionDetailWriter(BaseSingleWriter):
    model = TransactionDetail

    fields_config = [
        ('type_code', lambda w, tc: tc.code),
        'amount',
        ('funds_type', lambda w, ft: ft.value),
        ('availability', expand_availability),
        'bank_reference',
        'customer_reference',
        'text'
    ]

    def write(self):
        records = ['']
        fields = self._write_fields_from_config(self.fields_config)

        i = 0
        records[i] += self.model.code.value
        for field_name in fields:
            if field_name == 'text' and self.obj.text:
                text_cursor = 0
                if self.text_on_new_line:
                    records[i] += '/'
                    records.append(CONTINUATION_CODE)
                    i += 1

                while text_cursor < len(self.obj.text):
                    # -1 for comma after preceding field
                    remaining_line_length = (self.line_length - len(records[i])) - 1

                    if remaining_line_length > 0:
                        end_index = text_cursor + remaining_line_length
                        records[i] += ',' + self.obj.text[text_cursor:end_index]
                        text_cursor = end_index
                    else:
                        records.append(CONTINUATION_CODE)
                        i += 1
            else:
                records[i] += ',' + fields[field_name]

        return records


def expand_summary_items(writer, summary_items):
    items = OrderedDict()

    for n, summary_item in enumerate(summary_items):
        for summary_field_config in AccountIdentifierWriter.summary_fields_config:
            if isinstance(summary_field_config, str):
                summary_field_config = (summary_field_config, lambda w, x: x)

            summary_field_name, write_func = summary_field_config
            field_value = getattr(summary_item, summary_field_name, None)
            output = (write_func(writer, field_value)
                      if field_value is not None else None)

            if isinstance(output, dict):
                items.update(OrderedDict(
                    [('%s_%s' % (k, n), v) for k, v in output.items()]
                ))
            else:
                items['%s_%s' % (summary_field_name, n)] = convert_to_string(output)
    return items


class AccountIdentifierWriter(BaseSingleWriter):
    model = AccountIdentifier

    fields_config = [
        'customer_account_number',
        'currency',
        ('summary_items', expand_summary_items)
    ]

    summary_fields_config = [
        ('type_code', lambda w, tc: tc.code),
        'amount',
        'item_count',
        ('funds_type', lambda w, ft: ft.value),
        ('availability', expand_availability)
    ]

    def write(self):
        records = ['']
        fields = self._write_fields_from_config(self.fields_config)

        i = 0
        records[i] += self.model.code.value
        for field_name in fields:
            field_length = len(fields[field_name]) + 2
            if (len(records[i]) + field_length) >= self.line_length:
                records[i] += '/'
                records.append(CONTINUATION_CODE)
                i += 1
            records[i] += ',' + fields[field_name]
        records[i] += '/'
        return records


class AccountTrailerWriter(BaseSingleWriter):
    model = AccountTrailer

    fields_config = [
        'account_control_total',
        'number_of_records'
    ]


class AccountWriter(BaseSectionWriter):
    model = Account
    header_writer_class = AccountIdentifierWriter
    trailer_writer_class = AccountTrailerWriter
    child_writer_class = TransactionDetailWriter


class GroupHeaderWriter(BaseSingleWriter):
    model = GroupHeader

    fields_config = [
        'ultimate_receiver_id',
        'originator_id',
        ('group_status', lambda w, gs: gs.value),
        ('as_of_date', lambda w, d: write_date(d)),
        ('as_of_time', lambda w, t: write_time(t, w.clock_format_for_intra_day)),
        'currency',
        ('as_of_date_modifier', lambda w, aodm: aodm.value)
    ]


class GroupTrailerWriter(BaseSingleWriter):
    model = GroupTrailer

    fields_config = [
        'group_control_total',
        'number_of_accounts',
        'number_of_records'
    ]


class GroupWriter(BaseSectionWriter):
    model = Group
    header_writer_class = GroupHeaderWriter
    trailer_writer_class = GroupTrailerWriter
    child_writer_class = AccountWriter


class Bai2FileHeaderWriter(BaseSingleWriter):
    model = Bai2FileHeader

    fields_config = (
        'sender_id',
        'receiver_id',
        ('creation_date', lambda w, d: write_date(d)),
        ('creation_time', lambda w, t: write_time(t, w.clock_format_for_intra_day)),
        'file_id',
        'physical_record_length',
        'block_size',
        'version_number'
    )


class Bai2FileTrailerWriter(BaseSingleWriter):
    model = Bai2FileTrailer

    fields_config = (
        'file_control_total',
        'number_of_groups',
        'number_of_records',
    )


class Bai2FileWriter(BaseSectionWriter):
    model = Bai2File
    header_writer_class = Bai2FileHeaderWriter
    trailer_writer_class = Bai2FileTrailerWriter
    child_writer_class = GroupWriter



from .models import Record
from .constants import RecordCode


def _build_record(rows):
    fields_str = ''
    for index, row in enumerate(rows):
        field_str = row[1]

        if field_str[-1] == '/':
            fields_str += field_str[:-1] + ','
        else:
            fields_str += field_str + ' '

    fields = fields_str[:-1].split(',')
    return Record(code=rows[0][0], fields=fields, rows=rows)


def record_generator(lines):
    rows = iter(
        [(RecordCode(line[:2]), line[3:]) for line in lines]
    )

    records = [next(rows)]
    while True:
        try:
            row = next(rows)
        except StopIteration:
            break

        if row[0] != RecordCode.continuation:
            yield _build_record(records)
            records = [row]
        else:
            records.append(row)

    yield _build_record(records)


class IteratorHelper(object):
    def __init__(self, lines):
        self._generator = record_generator(lines)
        self.current_record = None
        self.advance()

    def advance(self):
        self.current_record = next(self._generator)
